/* min12xxw.c		driver for minolta pagepro 12xxW printers
 * 
 * It converts pages in pbmraw format read from stdin to the printer
 * language used by Minolta PagePro 1200W. The output goes to stdout.
 *
 * Copyright (C) 2004  Manuel Tobias Schiller <mala@hinterbergen.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/* Note: This program was written without any documentation from the
 * manufacturer. The description of the printer language may be
 * inaccurate or even wrong because it was obtained by looking at the
 * output of the windoof driver only. The code dealing with the
 * compression of raster data is in its spirit heavily based on an
 * initial driver by Adam Bocim <beetman@seznam.cz> who managed to
 * find out how things are done (thanks, Adam, you did a great job!).
 * This new version is considerably faster, implemented more cleanly
 * and (hopefully) well documented.
 * It also features querying the printer status and page counter, a
 * suggestion made by Bruno Schoedlbauer <bruno.schoedlbauer@gmx.de>,
 * who also pointed me to a nice USB sniffer by Benoit Papillault for
 * Windoof 98 and up, see http://benoit.papillault.free.fr/usbsnoop.
 * (This little program may prove very helpful on similar occasions.
 * Happy USB sniffing...)
 * */
/* revision history:
 * v. 0.0.0	(Jun  9 2004) initial revision based on Adam's driver
 * v. 0.0.1	(Jun 10 2004) fixed a bug when printing multipage
 * 		documents directly to the printer (the printer rejects
 * 		the data if the pauses between pages due to data processing
 * 		on the host are	too long, solution: buffer output in
 * 		temporary file)
 * v. 0.0.2	(Jun 14 2004) added correct treatment of margins
 * v. 0.0.3     (Jun 19 2004) added support for reading data from printer
 * 		as well as a toner-saving mode, fixed a bug in 300 dpi
 * 		output - this will be the version initially released to
 * 		the public
 * v. 0.0.4	(Jun 20 2004) made the raster data compression even
 *		faster and improved the compression ratio somewhat
 */

#define _GNU_SOURCE /* we need this because we use getline */
#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <getopt.h>

#ifndef MIN
#define MIN(x, y) (((x) < (y))?(x):(y))
#endif
#ifndef MAX /* not used now - but might come handy in the future */
#define MAX(x, y) (((x) > (y))?(x):(y))
#endif

static char *version = "version 0.0.4 (Jun 20 2004)";

/* defaults used for the output */
static int ptype = 0x00; /* normal paper */
static int pformat = 0x04; /* a4 */
static int res = 1; /* 600 dpi */
static int tray = 0xff; /* auto select paper tray */
static int nomargins = 0; /* if flag is set, margins are not enforced */
static int ecomode = 0; /* if set, toner-saving mode is enabled */
static char *device = "/dev/lp0"; /* default device used for queries */

/* try to exit gracefully in case of fatal errors */
static void fatal(char *str)
{
	perror(str);
	exit(1);
}

/**********************************************************************
 * minolta esc command utility
 *
 * this is used to send printer commands and read printer response
 * packets
 * see the file format.txt for an overview of what has been found out
 * about the format used
 **********************************************************************/
static void do_cmd(FILE *out, uint8_t cmd, uint32_t len, uint8_t *data)
{
	/* command sequence number, incremented with each command */
	static uint8_t sq = 0;
	uint8_t cksum = 0; /* every command is transmitted with a checksum */
	uint8_t buf[6]; /* buffer to build up command header */
	uint8_t *end = &data[len], *p = buf;

	assert(len < 0x10000); /* do not transmit way too much data */
	
	/* fill in command header */
	*p++ = 0x1b; *p++ = cmd; *p++ = sq++;
	*p++ = len & 0xff; *p++ = len >> 8; *p++ = ~cmd;
	
	/* calculate the checksum */
	p = data;
	while (p < end) cksum += *p++;
	p = buf;
	end = &buf[6];
	while (p < end) cksum += *p++;
	
	/* ok, write the command to the output file */
	if (fwrite(buf, sizeof(uint8_t), 6, out) != 6)
		fatal("min12xxw: error writing command to output file.");
	if (fwrite(data, sizeof(uint8_t), len, out) != len)
		fatal("min12xxw: error writing command to output file.");
	if (fputc(cksum, out) == EOF)
		fatal("min12xxw: error writing command to output file.");
}

/* this will read data from the printer */
static uint8_t *do_read(FILE *in, int *state, int *len)
{
	int c, i;
	uint8_t *retVal;

	if ((c = fgetc(in)) == EOF)
		fatal("min12xxw: error reading data from printer");
	if (state != NULL) *state = c;
	if ((c = fgetc(in)) == EOF)
		fatal("min12xxw: error reading data from printer");
	if (len != NULL) *len = c;

	if ((retVal = (uint8_t *) malloc(c * sizeof(uint8_t))) == NULL)
		fatal("min12xxw: error allocating buffer");
	for (i = 0; i < c; i++) {
		int cc;
		if ((cc = fgetc(in)) == EOF)
			fatal("min12xxw: error reading data from printer");
		retVal[i] = (uint8_t) cc;
	}
	
	return retVal;
}

/* sends the start signal to the printer (make printer pay attention) */
static void do_start(FILE *out)
{
	uint8_t buf[2];

	buf[0] = 0x81;
	buf[1] = 0;
	do_cmd(out, 0x40, 2, buf);
}

/* sends the stop signal to the printer (terminate a command sequence) */
static void do_stop(FILE *out)
{
	uint8_t buf = 0;

	do_cmd(out, 0x41, 1, &buf);
}

/* send query state command */
static void do_state(FILE *out, uint8_t page)
{
	uint8_t buf[2];

	buf[0] = page;
	buf[1] = 0;
	do_cmd(out, 0x60, 2, buf);
}

/* send query hardware info command */
static void do_hwinfo(FILE *out)
{
	uint8_t buf[3];

	buf[0] = 0x78;
	buf[1] = 0;
	buf[2] = 0x04;
	do_cmd(out, 0x6a, 3, buf);
}

/* read out page counter */
static uint32_t read_pagecnt(FILE *f)
{
	uint8_t *buf;
	int len, state;
	uint32_t retVal;
	
	do_start(f);
	do_hwinfo(f);
	buf = do_read(f, &state, &len); /* read in the first piece */
	if ((state != 0x45) || (len != 4))
		fatal("min12xxw: read unexpected data from printer");
	free(buf);
	buf = do_read(f, &state, &len); /* read in the second chunk */
	if ((state != 0x53) || (len != 38))
		fatal("min12xxw: read unexpected data from printer");
	retVal = buf[30] | (buf[31] << 8) | (buf[32] << 16) | (buf[33] << 24);
	free(buf);
	do_stop(f);
	
	return retVal;
}

/* read out status */
static char *read_status(FILE *f)
{
	uint8_t *buf;
	int len, state;
	
	do_start(f);
	do_state(f, 0x04); /* human readable status info is on page 4 */
	buf = do_read(f, &state, &len);
	if (state != 0x04)
		fatal("min12xxw: read unexpected data from printer");
	/* the string usually starts with a 0 byte and ends with none -
	 * make it start with useful data and zero-terminate it... */
	memmove(buf, &buf[1], len - 1);
	buf[len - 1] = 0;
	do_stop(f);
	
	return (char *) buf;
}

/* this procedure sends raster data to the printer */
static void send_raster_data(FILE *out, uint32_t nlines, \
				uint32_t len, uint8_t *data)
{
	uint8_t cmdbuf[6], *p = cmdbuf;

	assert(nlines < (1 << 16)); /* just to make sure */
	*p++ = len & 0xff; *p++ = (len >> 8) & 0xff;
	*p++ = (len >> 16) & 0xff; *p++ = len >> 24;
	*p++ = nlines & 0xff;
	*p++ = (nlines >> 8) & 0xff;
	/* we start with sending the command */
	do_cmd(out, 0x52, 6, cmdbuf);
	/* the data is appended immediately */
	if (fwrite(data, sizeof(uint8_t), len, out) != len)
		fatal("min12xxw: couldn't send raster data to output file");
}

/* this procedure sends a start-of-job/select-resolution-and-papertype
 * sequence to the printer */
static void send_start_job(FILE *out)
{
	uint8_t cmdbuf[8];

	memset(cmdbuf, 0x00, 8 * sizeof(uint8_t)); /* zero our buffer */

	
	do_start(out); /* start-of-printer-commands command */
	
	cmdbuf[0] = (uint8_t) res; /* set resolution and paper type */
	cmdbuf[3] = (uint8_t) ptype;
	cmdbuf[4] = 0x04;
	do_cmd(out, 0x50, 8, cmdbuf); /* issue it */
}

/* this procedure sends an end-of-job sequence to the printer */
static void send_end_job(FILE *out)
{
	uint8_t cmdbuf = 0;

	do_cmd(out, 0x55, 1, &cmdbuf);
	do_stop(out);
}

/* the following procedure sends a new page command */
static void send_new_page(FILE *out, uint32_t x, uint32_t y)
{
	uint8_t cmdbuf[22], *p = &cmdbuf[1];

	memset(cmdbuf, 0x00, 22 * sizeof(uint8_t)); /* zero our buffer */
	*p++ = 0x01;
	*p++ = (x >> 16) & 0xff; *p++ = x >> 24;
	*p++ = x & 0xff; *p++ = (x >> 8) & 0xff;
	*p++ = (y >> 16) & 0xff; *p++ = y >> 24;
	*p++ = y & 0xff; *p++ = (y >> 8) & 0xff;
	*p++ = 0x08; p++; *p++ = 0x08; p++;
	*p++ = tray; *p++ = pformat;
	if (res == 0) {
		/* apparently, 300 dpi needs special flags set here */
		cmdbuf[20] = 0xc0;
	}
	do_cmd(out, 0x51, 22, cmdbuf); /* issue our command */
}

/**********************************************************************
 * raster data compression
 **********************************************************************/
/* the following table keeps sixteen bytes which can be sent to the
 * printer by referring to their table index (4 bit only)
 * this table is filled anew for every scanline */
static uint32_t tbllen;
static uint8_t tbl[16];
static uint8_t invtbl[256]; /* this is for inverse lookups */

/* return the number of repetitions of a byte at position pointed to by p */
static uint32_t get_len(uint8_t *p, uint8_t *end)
{
	uint8_t c = *p;
	uint32_t len=0;
	
	while ((p < end) && (c == *p++)) len++;

	return len;
}

/* add a byte to the table and return the table index to it */
static uint8_t add_tbl(uint8_t b)
{
	/* check if the byte is already there */
	if (invtbl[b] < 16) return invtbl[b];
	/* check if there is space */
	if (tbllen >= 16) return 0xff;
	/* ok, add it to the table */
	tbl[tbllen] = b;
	invtbl[b] = tbllen;
	return tbllen++;
}

/* check if the next n bytes are in the table, or if we have space for them */
static int next_n_in_tbl(uint8_t *p, uint32_t len, uint8_t *end)
{
	uint32_t i = 0, j = len;
	
	/* make sure we don't read past the end of our data */
	if ((p + len) >= end) return 0;
	
	/* count how many bytes are in the table */
	while (j--)
		if (invtbl[*p++] < 16) i++;
	
	/* check if we have space for those which are not */
	if (tbllen  < (17 - len + i)) return 1;
	
	return 0; /* sorry - no space in table left... */
}

/* compress a scanline of data and write the result to an output buffer */
static uint32_t compress_scanline(uint8_t *p, uint8_t *end, uint8_t *obuf)
{
	uint32_t olen = 0;
	uint8_t *q; /* temporary pointer */
	
	tbllen = 0; /* clear the table and inverse lookup table */
	memset(invtbl, 0xff, 256 * sizeof(uint8_t));
	
	while (p < end) {
		uint32_t n = get_len(p, end);
		if (n > 2) {
			/* RLE compression pays off */
			/* make sure n is small enough - we're paranoid */
			assert(n < (63 * 64 + 63));
			if (n > 63) {
				/* how to encode multiples of 64 bytes */
				*obuf++ = 0xc0 | (n >> 6);
				*obuf++ = *p;
				olen += 2;
				p += n & (~0x3f);
				n &= 0x3f;
			}
			if (n) {
				/* encode the remainder */
				*obuf++ = 0x80 | n;
				*obuf++ = *p;
				olen += 2;
				p += n;
			}
		} else if (next_n_in_tbl(p, 4, end)) {
			/* ok - try to do table compression */
			/* the next four bytes are in the table or there
			 * is enough space to put them into the table
			 * we issue a 0x41 code to indicate that we send
			 * table indices - the 0x41 means that we send
			 * a quartet of table indices - we might even
			 * send more, see below */
			q = obuf; /* back up current position */
			*obuf++= 0x41;
			*obuf = add_tbl(*p++) << 4;
			*obuf++ |= add_tbl(*p++);
			*obuf = add_tbl(*p++) << 4;
			*obuf++ |= add_tbl(*p++);
			olen += 3;
			/* check if we can send even more table indices */
			while (next_n_in_tbl(p, 2, end) && (*q < 0x7f)) {
				/* break if we can RLE-code the next bytes */
				if (get_len(p, MIN(p + 3, end))	>= 3) break;
				/* we send one more pair of table indices */
				(*q)++;
				*obuf = add_tbl(*p++) << 4;
				*obuf++ |= add_tbl(*p++);
				olen++;
			}
		} else {
			/* can't do table compression - fall back to send
			 * up to 10 bytes in uncompressed plaintext */
			q = obuf; /* back up current output position */
			*obuf++ = 0xff;
			olen++;
			do {
				*obuf++ = *p++;
				olen++;
				(*q)++; /* update # of plaintext bytes */
				/* three repeating bytes ahead? if so,
				 * do RLE compression instead */
				if (get_len(p, MIN(p + 3, end)) >= 3) break;
				/* see if we can do table compression for
				 * the bytes ahead - comment the next line
				 * of code out if you absolutely need the
				 * speed and can live with things compressing
				 * a little worse (on the order of a percent
				 * or so) */
				if (next_n_in_tbl(p, 4, end)) break;
			} while ((p < end) && (*q < 9));
		}
	}
	
	return olen;
}

/********************************************************************** 
 * process all the pages in the rawpbm input job
 **********************************************************************/
/* printer has non-printable margins of 17/100" on all sides of a sheet
 * of paper which are present in ghostscript output and need to be
 * removed - these are in units of 8 pixels; note - this is not exact
 * to the last decimal place but it has to be good enough
 * update this if new printers allow for new resolutions codes or equal
 * codes with different meanings!!! */
static int skiptbl[3] = { 6, 13, 25 };

/* processes a single page */
static void dopage(FILE *in, FILE *out, uint32_t x, uint32_t y)
{
	uint32_t sclbytes = x / 8; /* # of bytes per scanline */
	/* # of scanlines per block (rounded up) */
	uint32_t sclperbl = y / 8 + ((y & 7)?1:0);
	/* size of buffer for processed data - note: the estimate given
	 * below is conservative - the buffer can not become larger... */
	uint32_t blblen = sclperbl * (17 + sclbytes + sclbytes / 10 + 1);
	uint8_t *scl, *blbuf; /* buffers for scanline and output block */
	uint32_t bllen; /* block length so far */
	uint32_t sclen; /* length of scanline excluding table */
	uint32_t yy, yc; /* counters */
	/* this controls how many pixels are skipped on each side of a
	 * sheet of paper to keep the margins that the printer expects */
	int skip = ((nomargins == 0)?skiptbl[res]:0);
	int i; /* counter */
	int ecofl = 0; /* flag used for toner-saving mode */
	
	/* allocate buffers */
	if ((scl = (uint8_t *) malloc(sclbytes * sizeof(uint8_t))) == NULL)
		fatal("min12xxw: couldn't allocate memory");
	if ((blbuf = (uint8_t *) malloc(blblen * sizeof(uint8_t))) == NULL)
		fatal("min12xxw: couldn't allocate memory");
	
	yc = 0;
	for (yc = 0; yc < skip * 8; yc++) {
		/* skip the first few scanlines in case we enforce margins */
		if (fread(scl, sizeof(uint8_t), sclbytes, in) != sclbytes)
			fatal("min12xxw: couldn't read scanline");
	}
	/* we will always build eight blocks per page - the windows driver
	 * seems to do the same thing */
	for (i = 0; i < 8; i++) {
		bllen = 0;
		/* process the scanlines in the block */
		for (yy = 0; (yy < sclperbl) && (yc < y); yy++, yc++) {
			/* read in the next scanline */
			if (fread(scl, sizeof(uint8_t), sclbytes, in) != \
				sclbytes)
				fatal("min12xxw: couldn't read scanline");
			/* check if we need to do this line (we don't work
			 * for the margins...) */
			if ((yc + skip * 8) > y) continue;
			/* check if in toner-saving mode */
			if (ecomode) {
				/* clear every other scanline */
				if (ecofl) memset(scl, 0,
						sclbytes * sizeof(uint8_t));
				ecofl = ~ecofl;
			}
			/* compress a scanline - leave some space for the
			 * table (we'll probably have to remove a hole) */
			sclen = compress_scanline(&scl[skip], \
						&scl[sclbytes-skip], \
						blbuf + 17);
			blbuf[0] = 0x80 + tbllen; /* copy table */
			if (tbllen) memcpy(blbuf + 1, tbl, tbllen);
			if (tbllen < 16) /* remove the hole */
				memmove(blbuf + 1 + tbllen, blbuf + 17, sclen);
			blbuf += sclen + 1 + tbllen;
			bllen += sclen + 1 + tbllen;
		}
		blbuf -= bllen;
		/* account for skipped scanlines in the last block */
		if (i == 7) yy -= 8 * skip;
		/* send the raster data to the printer */
		send_raster_data(out, yy, bllen, blbuf);
	}
	free(scl); /* avoid memory leaks */
	free(blbuf);
}

/* processes all the pages in a pbmraw job file */
static void dojob(FILE *in, FILE *out)
{
	uint32_t x, y; /* page dimesions in dots at the current resolution */
	char *line = NULL; /* buffer for reading whole lines from the input */
	size_t llen = 0; /* length of line buffer */
	/* how much to skip because of margins */
	int skip = ((nomargins == 0)?(16 * skiptbl[res]):0);
	
	/* send the printer a start-of-job/set-resolution-and-papertype
	 * sequence */
	send_start_job(out);

	/* process pages */
	while (!feof(in) && !ferror(in)) {
		/* read the png header for x and y dimensions */
		if (getline(&line, &llen, in) == -1) break;
		/* check for valid pbmraw signature */
		if ((line[0] != 'P') || (line[1] != '4'))
			fatal("min12xxw: input is not valid pbmraw");
		/* read in comment lines and the line containing the
		 * resolution */
		do {
			if (getline(&line, &llen, in) == -1)
				fatal("min12xxw: input is not valid pbmraw");
		} while (line[0] == '#');
		/* parse in resolution */
		if (sscanf(line, "%lu %lu", (unsigned long *) &x, \
				(unsigned long *) &y) != 2)
			fatal("min12xxw: input is not valid pbmraw");
		/* scanlines are byte-aligned - adjust x accordingly */
		if (x & 0x7) x = 8 + (x & ~0x7);
		
		/* tell the printer to start a new page - note the page
		 * dimensions are reduced because we need to take the
		 * margins into account - see above */
		send_new_page(out, x - skip, y - skip);
		/* read, process and send the raster data for the next
		 * page  - this needs to know the real dimensions because
		 * it needs to know how much data to read... */
		dopage(in, out, x, y);
	}

	/* tell the printer we got our job finished... */
	send_end_job(out);
	
	free(line); /* free memory to avoid leaks... */
}

/**********************************************************************
 * option handling
 **********************************************************************/
/* this procedure gives help */
static void help()
{
	fprintf(stderr, "min12xxw - a pbmraw to Milolta PagePro 12xx W " \
			"filter\n\noptions: (defaults are marked with an" \
			" asterisk)\n" \
			"\t-h\t--help\t\t\tthis help\n" \
			"\t-v\t--version\t\tshow version number\n" \
			"\t-c\t--count\t\t\tquery printer page count\n" \
			"\t-d dev\t--device dev\t\tset device to use for " \
			"queries\n" \
			"\t-s\t--status\t\tquery printer status\n" \
			"\t-e\t--ecomode\t\teconomic (toner saving) mode\n" \
			"\t-n\t--nomargins\t\tdisable enforcement of " \
			"margins\n" \
			"\t-r res\t--res res\t\tset resolution\n" \
			"\t-t tray\t--tray tray\t\tset paper tray\n" \
			"\t-p type\t--papertype type\tset paper type" \
			"\n\t-f fmt\t--paperformat fmt\tset paper format\n" \
			"\navailable resolutions in dpi: 300, 600*, 1200\n" \
			"\navailable paper trays: auto*, tray1, tray2, " \
			"manual\n\navailable paper types: normal*, thick, "\
			"envelope, postcard\n" \
			"\navailable paper formats: " \
			"a4*, b5, a5, jpost, corpost,\njisy6, jisy0, "\
			"chinese16k, chinese32k, legal, glegal,\n" \
			"letter, gletter, executive, halfletter, " \
			"envmonarch,\nenv10, envdl, envc5, envc6," \
			"envb5, choukei3gou, choukei5gou,\ncustom, " \
			"envb6, folio, jisy1, jisy2, quadpost\n\n");
	exit(1);
}

/* the next few tables contain the mappings from human-readable strings
 * to id numbers the printer understands */
struct map {
	char *str;
	int id;
};
static struct map ptypes[] = {
	{ "normal", 0x00 }, { "thick", 0x01 }, { "transparency", 0x02 },
	{ "postcard", 0x03 }, { "envelope", 0x03 }, { NULL, -1 }
};
static struct map trays[] = {
	{ "auto", 0xff }, { "tray1", 0x00 }, { "tray2", 0x01 },
	{ "manual", 0x80 }, { NULL, -1 }
};
static struct map pformats[] = {
	{ "a4", 0x04 }, { "b5", 0x06 }, { "a5", 0x08 }, { "jpost", 0x0c },
	{ "corpost", 0x0d }, { "jisy6", 0x10 }, { "jisy0", 0x11 },
	{ "chinese16k", 0x13 }, { "chinese32k", 0x15 }, { "legal", 0x19 },
	{ "glegal", 0x1a }, { "letter", 0x1b }, { "gletter", 0x1d },
	{ "executive", 0x1f }, { "halfletter", 0x21 }, { "envmonarch", 0x24 },
	{ "env10", 0x25 }, { "envdl", 0x26 }, { "envc5", 0x27 },
	{ "envc6", 0x28 }, { "envb5", 0x29 }, { "choukei3gou", 0x2d },
	{ "choukei5gou", 0x2e }, { "custom", 0x31 }, { "envb6", 0x31 },
	{ "folio", 0x31 }, { "jisy1", 0x31 }, { "jisy2", 0x31 },
	{ "quadpost", 0x31 }, { NULL, -1 }
};

/* convert string to id number */
static int getid(struct map *m, char *thing, char *str, int deflt)
{
	while (m->str != NULL) {
		if (strcasecmp(str, m->str) == 0) return m->id;
		m++;
	}
	fprintf(stderr, "min12xxw: unknown %s: %s\n", thing, str);
	return deflt; /* return the default */
}

/* convert resolution to id number */
static int getres(unsigned char *res)
{
	int r;

	if (sscanf(res, "%d", &r) == 1)	{
		switch (r) {
		case 300:
			return 0;
		case 600:
			return 1;
		case 1200:
			return 2;
		}
	}
	fprintf(stderr, "min12xxw: unknown resolution: %s\n", res);
	return 1;
}

/* parse the options given to the filter */
static void parseopts(int argc, char **argv)
{
	/* this structure is needed by getopt_long */
	static struct option opts[] = {
		{ "res",		required_argument, 0, 'r' },
		{ "papertype",		required_argument, 0, 'p' },
		{ "paperformat",	required_argument, 0, 'f' },
		{ "tray",		required_argument, 0, 't' },
		{ "nomargins",		no_argument,	   0, 'n' },
		{ "ecomode",		no_argument,	   0, 'e' },
		{ "version",		no_argument,	   0, 'v' },
		{ "device",		required_argument, 0, 'd' },
		{ "count",		no_argument,	   0, 'c' },
		{ "status",		no_argument,	   0, 's' },
		{ "help",		no_argument,	   0, 'h' },
		{ NULL,			0,		   0,   0 }
	};
	int optidx = 0, dopgcnt = 0, dostate = 0;
	int c;

	while (42) { /* this line is for all hitch-hikers out there */
		c = getopt_long(argc, argv, "hcsevnd:r:p:f:t:", opts, &optidx);
		if (c == -1) break;
		switch (c) {
		case 'h':
			help();
		case 'v':
			printf("min12xxw: %s\n", version);
			exit(0);
		case 'r':
			res = getres(optarg);
			break;
		case 'p':
			ptype = getid(ptypes, "paper type", optarg, 0);
			break;
		case 'f':
			pformat = getid(pformats, "paper format", optarg, 4);
			break;
		case 't':
			tray = getid(trays, "tray", optarg, 0xff);
			break;
		case 'n':
			nomargins = 1;
			break;
		case 'd':
			device = optarg;
			break;
		case 'c':
			dopgcnt = 1;
			break;
		case 's':
			dostate = 1;
			break;
		case 'e':
			ecomode = 1;
			break;
		default:
			fprintf(stderr, "min12xxw: unknown option: -%c\n", c);
			exit(1);
		}
	}
	if ((dopgcnt) || (dostate)) {
		/* process printer queries right away and exit */
		FILE *d = fopen(device, "r+b"); /* open our printer */
		if (d == NULL)
			fatal("min12xxw: couldn't open printer");
		if (dopgcnt) printf("min12xxw: %s reports %lu pages\n", \
			device, (unsigned long) read_pagecnt(d));
		if (dostate) {
			char *s = read_status(d);
			printf("min12xxw: %s reports state: %s\n", device, s);
			free(s);
		}
		fclose(d); /* close the printer */
		exit(0); /* exit */
	}
}

/**********************************************************************
 * main program
 **********************************************************************/
int main(int argc, char **argv)
{
	FILE *tmp;
	long i;
		
	/* this is really boring now... */
	parseopts(argc, argv); /* parse any options given */
	/* create temporary file to hold the output until we're done */
	if ((tmp = tmpfile()) == NULL)
		fatal("min12xxw: couldn't create temporary file");
	dojob(stdin, tmp); /* filter our input */
	/* copy output from temp file to stdout */
	rewind(tmp);
	while (!feof(tmp)) {
		char buf[4096];
		i = fread(buf, 1, 4096, tmp);
		if (i <= 0) continue;
		if (fwrite(buf, 1, i, stdout) != i)
			fatal("min12xxw: couldn't write to stdout");
	}

	return 0;
}
